MODULE SunRiseSet;
(* 
  Example taken from williams.best.vwh.net.sunrise_sunset_example.htm,
  article of algorithm at williams.best.vwh.sunrise_sunset_algorithm.htm 
*)
IMPORT Math0, Reals, Out;

VAR
  zenith : REAL; (* official Zenith, see article *)
  
  (* input var *)
  longitude, latitude : REAL;
  (* calc var *)
  lngHour : REAL;
  t : REAL;
  SunLongitude : REAL;
  M : REAL;    (* Sun's anomaly *)
  RA : REAL;   (* Right Ascension *)
  sinDec, cosDec : REAL;
  H, T, UT : REAL;
  N : INTEGER; (* Number of days *)
  SunRise, info : BOOLEAN;

PROCEDURE WrTR(txt : ARRAY OF CHAR;v : REAL);
VAR
  aoc  :ARRAY 20 OF CHAR;
BEGIN
  IF info THEN
    Reals.RealToStrF(v, 7, aoc);
    Out.String(txt); Out.String(aoc); Out.Ln;
  END    
END WrTR;

PROCEDURE DegreeAdjust(VAR v : REAL);
BEGIN
  IF v > 360.0 THEN v := v - 360.0
  ELSIF v < 0.0 THEN v := v + 360.0
  END
END DegreeAdjust;

PROCEDURE DayAdjust(VAR v : REAL);
BEGIN
  IF v > 24.0 THEN v := v - 24.0
  ELSIF v < 0.0 THEN v := v + 24.0
  END
END DayAdjust;

PROCEDURE FNday (year, month, day : INTEGER): INTEGER;
VAR
  N1, N2, N3 : INTEGER;
BEGIN
  N1 := 275 * month DIV 9;
  N2 := (month + 9) DIV 12;
  N3 := (1 + ((year - 4 * (year DIV 4) + 2) DIV 3));
  RETURN N1 - (N2 * N3) + day - 30
END FNday;

PROCEDURE Step2;
BEGIN
  lngHour := longitude / 15.0;
  WrTR("Step2, lngHour=", lngHour);  
  IF SunRise THEN
    t := FLT(N) + (( 6.0 - lngHour) / 24.0)   (* for SunRise *)
  ELSE
    t := FLT(N) + (( 18.0 - lngHour) / 24.0)  (* for SunSet *)
  END;
  WrTR("Step2, t=", t);  
END Step2;

PROCEDURE Step3;
BEGIN
  M := (0.9856 * t) - 3.289;
  WrTR("Step3, M=", M);  
END Step3;

PROCEDURE Step4;
BEGIN
  SunLongitude := M + (1.916 * Math0.DegSin(M)) + (0.020 * Math0.DegSin(2.0 * M)) + 282.634;
  DegreeAdjust(SunLongitude);
  WrTR("Step4, L=", SunLongitude);  
END Step4;

PROCEDURE Step5;
VAR
  Lquadrant, RAquadrant : REAL;
BEGIN
  RA := Math0.DegArcTan(0.91764 * Math0.DegTan(SunLongitude));
  WrTR("Step5 DegTan=", Math0.DegTan(SunLongitude));
  WrTR("Step5a, RA=", RA);
  DegreeAdjust(RA);
  Lquadrant := FLT(FLOOR(SunLongitude / 90.0)) * 90.0;
  WrTR("Step5b, Lquadrant=", Lquadrant);  
  RAquadrant := FLT(FLOOR(RA / 90.0)) * 90.0;
  WrTR("Step5b, RAquadrant=", RAquadrant);  
  RA := (RA + (Lquadrant - RAquadrant)) / 15.0; (* right ascension value *)
  WrTR("Step5b, RA=", RA);  
END Step5;

PROCEDURE Step6;
BEGIN
  sinDec := 0.39782 * Math0.DegSin(SunLongitude);
  WrTR("Step6, sinDec=", sinDec);  
  cosDec := Math0.DegCos(Math0.DegArcSin(sinDec));
  WrTR("Step6, cosDec=", cosDec);  
END Step6;

(* output : Always in range 0, 1, or 2 *)
PROCEDURE Step7(): INTEGER;
VAR
  note : INTEGER;
  cosH : REAL;
BEGIN
  cosH := (Math0.DegCos(zenith) - (sinDec * Math0.DegSin(latitude))) / (cosDec * Math0.DegCos(latitude));
  WrTR("Step7, cosH=", cosH);  
  IF cosH > 1.0 THEN
    note := 1
  ELSIF cosH < -1.0 THEN
    note := 2
  ELSE
    IF SunRise THEN
      H := 360.0 - Math0.DegArcCos(cosH)  (* for SunRise *)
    ELSE
      H := Math0.DegArcCos(cosH)   (* for SunSet  *)
    END;
    H := H / 15.0;
    note := 0;
    WrTR("Step7, H=", H);    
  END;
  RETURN note
END Step7;

PROCEDURE Step8;
BEGIN
  T := H + RA - (0.06571 * t) - 6.622;
  WrTR("Step8, T=", T);  
END Step8;

PROCEDURE Step9;
BEGIN
  UT := T - lngHour;
  WrTR("Step9, UT=", UT);
  DayAdjust(UT);
  WrTR("Step9, Corrected UT=", UT);  
END Step9;

(* All parameters should contain sensible values, 
   On exit, check for negative values for error-conditions:
    >=0 = a valid number in real format, so must be converted to hh.mm notation; see testcase
     -1 = The given location does not have a sunrise 
     -2 = The given location does not have a sunset  *)
     
PROCEDURE GetTime*(MyLatitude, MyLongitude, MyTimeZone : REAL; RiseOrSet : BOOLEAN; y, m, d : INTEGER): REAL;
VAR
  rv : REAL;
BEGIN
  latitude :=  MyLatitude; longitude := MyLongitude; SunRise := RiseOrSet;
  WrTR("Latitude=", latitude); WrTR("Longitude=", longitude);  
  N := FNday(y, m, d);
  WrTR("Step1, N=", FLT(N));
  Step2; Step3; Step4; Step5; Step6;
  CASE Step7() OF
    0:  Step8; Step9; rv := UT + MyTimeZone; DayAdjust(rv)  (* local offset adjustment *)
  | 1 : rv := -1.0 (* sun never rises *)
  | 2 : rv := -2.0 (* sun never sets *)
  END;
  RETURN rv
END GetTime;


BEGIN
  zenith := 90.833; (* offical Sun's Zenith *)
  info := FALSE
END SunRiseSet.
